pub enum Status {
SATISFIABLE = 10,
UNSATISFIABLE = 20,
UNKNOWN = 0,
}Expand description
The SAT competition standardized the exit code of SAT solvers to the following which then is also used return code for ‘solve’ functions. In the following example we use those constants for brevity though.
Variants