pub trait Learner {
// Required methods
fn learning(&mut self, size: i32) -> bool;
fn learn(&mut self, lit: i32);
}Expand description
Connected learners which can be used to export learned clauses. The ‘learning’ can check the size of the learn clause and only if it returns true then the individual literals of the learned clause are given to the learn through ‘learn’ one by one terminated by a zero literal.