Terminator

Trait Terminator 

pub trait Terminator {
    // Required method
    fn terminated(&mut self) -> bool;
}
Connected terminators are checked for termination regularly. If the ‘terminate’ function of the terminator returns true the solver is terminated synchronously as soon it calls this function.

Required Methods

fn terminated(&mut self) -> bool

Implementors